package eu.jonahbauer.json.query.parser.ast;

import eu.jonahbauer.json.JsonNumber;
import eu.jonahbauer.json.JsonValue;
import eu.jonahbauer.json.query.JsonMath;
import eu.jonahbauer.json.query.JsonQueryException;
import eu.jonahbauer.json.query.impl.Generator;
import eu.jonahbauer.json.query.parser.ast.JQExpression.Context;
import eu.jonahbauer.json.query.util.Util;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Collectors;

public record JQPatterns(@NotNull List<@NotNull Pattern> patterns) {

    public JQPatterns {
        patterns = List.copyOf(patterns);
        if (patterns.isEmpty()) throw new IllegalArgumentException();
    }

    public @NotNull Generator<@Nullable JsonValue> bind(@NotNull Context context, @Nullable JsonValue value, Function<@NotNull Context, @NotNull Generator<@Nullable JsonValue>> downstream) {
        return new PatternGeneratorImpl(patterns, context, value, downstream);
    }

    private static final class PatternGeneratorImpl implements Generator<@Nullable JsonValue> {
        // one generator per pattern
        private final @NotNull Generator<@NotNull Generator<@NotNull Map<@NotNull String, @Nullable JsonValue>>> patterns;
        private final @NotNull Context context;
        private final @NotNull Function<@NotNull Context, @NotNull Generator<@Nullable JsonValue>> downstream;

        // the generator for the current pattern
        private @Nullable Generator<@Nullable JsonValue> current;

        public PatternGeneratorImpl(
                @NotNull List<@NotNull Pattern> patterns, @NotNull Context context, @Nullable JsonValue value,
                @NotNull Function<@NotNull Context, @NotNull Generator<@Nullable JsonValue>> downstream
        ) {
            this.context = context;
            this.downstream = downstream;
            var variables = new HashMap<String, JsonValue>();
            patterns.stream().map(Pattern::variables).flatMap(Collection::stream).forEach(key -> variables.put(key, null));
            this.patterns = Generator.from(patterns).map(pattern -> pattern.bind(context, value).map(vars -> {
                var out = new HashMap<>(variables);
                out.putAll(vars);
                return out;
            }));
        }

        private <T> T advance(@NotNull Function<Generator<@Nullable JsonValue>, T> function) {
            var ex = new JsonQueryException("no match");
            while (true) {
                if (current == null) {
                    // find the next matching pattern
                    while (patterns.hasNext()) {
                        try {
                            var match = patterns.next();
                            current = match.map(context::withVariables).flatMap(downstream);
                            break;
                        } catch (JsonQueryException e) {
                            ex = e;
                        }
                    }

                    // no matching pattern; propagate exception from last match attempt
                    if (current == null) {
                        throw ex;
                    }
                }

                try {
                    return function.apply(current);
                } catch (JsonQueryException e) {
                    // exception during execution; try the next pattern
                    current = null;
                    ex = e;
                }
            }
        }

        @Override
        public boolean hasNext() {
            return advance(Generator::hasNext);
        }

        @Override
        public @Nullable JsonValue next() throws NoSuchElementException {
            return advance(Generator::next);
        }
    }

    @Override
    public @NotNull String toString() {
        return patterns.stream().map(Objects::toString).collect(Collectors.joining(" ?// "));
    }

    public sealed interface Pattern {
        @NotNull
        Set<@NotNull String> variables();
        @NotNull
        Generator<Map<@NotNull String, @Nullable JsonValue>> bind(@NotNull Context context, @Nullable JsonValue value);

        record ValuePattern(@NotNull String name) implements Pattern {
            public ValuePattern {
                Objects.requireNonNull(name);
                if (!name.startsWith("$")) throw new IllegalArgumentException();
            }

            @Override
            public @NotNull Set<@NotNull String> variables() {
                return Set.of(name);
            }

            @Override
            public @NotNull Generator<Map<@NotNull String, @Nullable JsonValue>> bind(@NotNull Context context, @Nullable JsonValue value) {
                var map = new HashMap<String, JsonValue>();
                map.put(name, value);
                return Generator.of(Collections.unmodifiableMap(map));
            }

            @Override
            public @NotNull String toString() {
                return name;
            }
        }

        record ArrayPattern(@NotNull List<@NotNull Pattern> patterns) implements Pattern {
            public ArrayPattern {
                patterns = List.copyOf(patterns);
                if (patterns.isEmpty()) throw new IllegalArgumentException();
            }

            @Override
            public @NotNull Set<@NotNull String> variables() {
                var out = new HashSet<String>();
                patterns.forEach(p -> out.addAll(p.variables()));
                return Collections.unmodifiableSet(out);
            }

            @Override
            public @NotNull Generator<Map<@NotNull String, @Nullable JsonValue>> bind(@NotNull Context context, @Nullable JsonValue value) {
                var streams = new ArrayList<Supplier<Generator<Map<String, JsonValue>>>>();

                for (int i = 0; i < patterns.size(); i++) {
                    var k = new JsonNumber(i);
                    var v = JsonMath.index(value, k);

                    var pattern = patterns.get(i);
                    streams.add(() -> pattern.bind(context, v));
                }

                return Util.crossReversed(streams).map(list -> {
                    var map = new HashMap<String, JsonValue>();
                    list.forEach(map::putAll);
                    return map;
                });
            }

            @Override
            public @NotNull String toString() {
                return patterns.stream().map(Objects::toString).collect(Collectors.joining(", ", "[", "]"));
            }
        }

        record ObjectPattern(@NotNull SequencedMap<@NotNull JQExpression, @NotNull Pattern> patterns) implements Pattern {

            public ObjectPattern {
                patterns = Collections.unmodifiableSequencedMap(new LinkedHashMap<>(patterns));
                if (patterns.isEmpty()) throw new IllegalArgumentException();
            }

            @Override
            public @NotNull Set<@NotNull String> variables() {
                var out = new HashSet<String>();
                patterns.values().forEach(p -> out.addAll(p.variables()));
                return Collections.unmodifiableSet(out);
            }

            @Override
            public @NotNull Generator<Map<@NotNull String, @Nullable JsonValue>> bind(@NotNull Context context, @Nullable JsonValue value) {
                var streams = new ArrayList<Supplier<Generator<Map<String, JsonValue>>>>();

                this.patterns.reversed().forEach((keyExpression, pattern) -> streams.add(() -> {
                    var keyStream = keyExpression.evaluate(context);
                    return keyStream.flatMap(key -> pattern.bind(context, JsonMath.index(value, key)));
                }));

                return Util.cross(streams).map(list -> {
                    var map = new HashMap<String, JsonValue>();
                    list.reversed().forEach(map::putAll);
                    return map;
                });
            }

            @Override
            public @NotNull String toString() {
                var out = new StringJoiner(", ", "{", "}");
                patterns.forEach((key, pattern) -> out.add(key + ": " + pattern));
                return out.toString();
            }
        }
    }
}
